import pytest

from vesna import providers


@pytest.mark.parametrize(
    "provider_cls, filename",
    [
        (providers.ProviderJSON, "uk.json"),
        (providers.ProviderJSON5, "uk.json5"),
        (providers.ProviderHJSON, "uk.hjson"),
        (providers.ProviderYAML, "uk.yaml"),
        (providers.ProviderTOML, "uk.toml"),
    ],
)
async def test_provider_load_save(provider_cls, filename, tmp_path):
    file_path = tmp_path / filename

    provider = provider_cls()

    await provider.load_file("tests/localisation/" + filename, "uk")
    await provider.save_file(file_path)

    await provider.load_file(file_path, "uk")
    assert provider.get("hello") == "Привіт {}"

    provider.set("hello", "Здоров {}")
    await provider.save_file()

    await provider.load_file(file_path, "uk")
    assert provider.get("hello") == "Здоров {}"



import asyncio

import vesna


async def main():
    await vesna.i18n.load_file(
        vesna.providers.ProviderJSON(), "uk", "tests/localisation/{locale_code}.json"
    )

    l = vesna.Locale("uk")  # noqa: E741
    print(l["vesna"])
    print(l("hello", l.get("world")))


if __name__ == "__main__":
    asyncio.run(main())



import pathlib

import aiofiles
import pydantic
from pydantic_core import from_json

from vesna.providers.dict_provider import DataDictModel, DictProvider


class ProviderJSON(DictProvider):
    """
    Implementation of IO methods for DictProvider to support the JSON format.
    Uses pydantic_core as a backend.
    """

    async def load_file(self, path: pathlib.Path, locale_code: str) -> None:
        await self.clean()

        async with aiofiles.open(path, "r", encoding="utf-8") as f:
            content = await f.read()

            try:
                self._storage = JSONDataModel.model_validate_json(content)
            except pydantic.ValidationError:
                raw_dict = from_json(content)

                self._storage = JSONDataModel(
                    data=raw_dict,
                    locale_code=locale_code,
                )

            self._storage.path = path

    async def save_file(self, path: pathlib.Path | None = None) -> None:
        if self._storage is None:
            raise ValueError("The storage is None")

        path = pathlib.Path(path) if path else self._storage.path
        if not path:
            raise ValueError(f"Path cannot be None | Path: {path}")

        async with aiofiles.open(path, "w", encoding="utf-8") as f:
            await f.write(self._storage.model_dump_json())


class JSONDataModel(DataDictModel):
    pass



import pathlib
import typing
from typing import Protocol


class ProviderProtocol(Protocol):
    """
    Provides a basic protocol for providers that will be compatible with Vesna.
    """

    @classmethod
    async def from_file(cls, path: pathlib.Path, locale_code: str) -> ProviderProtocol: ...

    def __getitem__(self, key: str) -> str | None: ...

    def __setitem__(self, key: str, value: str) -> str: ...

    def get(self, key: str, default: str | None = None, **kwargs) -> str | None: ...

    def set(self, key: str, value: str) -> str: ...

    def get_storage(self) -> ModelDataProtocol: ...

    def get_locale_code(self) -> str | None: ...

    def get_file_path(self) -> pathlib.Path | None: ...

    async def load_file(self, path: pathlib.Path, locale_code: str) -> None: ...

    async def save_file(self, path: pathlib.Path | None = None) -> None: ...

    async def clean(self) -> None: ...

    def is_empty(self) -> bool: ...


class ModelDataProtocol(Protocol):
    """
    Provides a base protocol for a model that will be compatible with ProviderProtocol.
    """

    data: typing.Any
    locale_code: str | None  # ISO 15897 (ISO 639-1 + ISO 3166-1 alpha-2)

    path: pathlib.Path | None = None

    version: str



import datetime
import pathlib

import pytest

from vesna import Locale, Vesna, vesna
from vesna.providers import FluentProvider, GettextProvider, ProviderJSON


@pytest.fixture
def temp_json_locale(tmp_path: pathlib.Path) -> pathlib.Path:
    file_path = tmp_path / "uk.json"
    # Nested structure to match DictProvider dot-notation behavior
    file_path.write_text(
        '{"greeting": "Привіт, {}!", "apples": {"one": "{count} яблуко", "few": "{count} яблука", "many": "{count} яблук", "other": "{count} яблука"}}',  # noqa: E501
        encoding="utf-8",
    )
    return tmp_path


async def test_vesna_fallback(temp_json_locale):
    """Checking fallback logic and error handling."""
    i18n = Vesna()
    i18n.providers = {}
    i18n.default_locale = "uk"

    await i18n.load_file(ProviderJSON(), "uk", temp_json_locale / "uk.json")

    en_path = temp_json_locale / "en.json"
    en_path.write_text('{"other": "Something else"}', encoding="utf-8")
    await i18n.load_file(ProviderJSON(), "en", en_path)

    assert i18n.get_text("greeting", "en") == "Привіт, {}!"
    assert i18n.get_text("not_exist", "en") == "not_exist"

    with pytest.raises(KeyError):
        i18n.get_text("not_exist", "en", is_exception=True)


async def test_locale_babel_formatting(temp_json_locale):
    """Verifying integration with Babel: dates, plural forms, currency."""
    i18n = Vesna()
    i18n.providers = {}
    vesna.providers = {}

    await i18n.load_file(ProviderJSON(), "uk", temp_json_locale / "uk.json")

    locale = Locale("uk", vesna=i18n)

    dt = datetime.datetime(2026, 3, 10)
    formatted_date = locale.date(dt)
    assert "2026" in formatted_date

    assert locale.currency(100.50, "UAH") == "100,50\xa0₴"
    assert locale.number(1234.56) == "1\xa0234,56"

    # Plural keys: apples.one, apples.few, etc.
    assert locale.plural("apples", 1) == "1 яблуко"
    assert locale.plural("apples", 2) == "2 яблука"
    assert locale.plural("apples", 5) == "5 яблук"


async def test_fluent_provider(tmp_path: pathlib.Path):
    """Checking the Fluent provider and its specific formatting."""
    ftl_path = tmp_path / "uk.ftl"
    ftl_path.write_text(
        "hello-user = Привіт, { $userName }!\nmissing-var = Дані: { $data }", encoding="utf-8"
    )

    provider = FluentProvider(use_isolating=False)
    await provider.load_file(ftl_path, "uk")

    assert provider.get("hello-user", userName="Олексій") == "Привіт, Олексій!"
    assert provider.get("not-exist", default="Дефолт") == "Дефолт"

    with pytest.raises(RuntimeError):
        provider.get("missing-var")


async def test_gettext_provider_po(tmp_path: pathlib.Path):
    """Checking the Gettext (.po) provider."""
    po_content = """
    msgid ""
    msgstr ""
    "Language: uk\\n"

    msgid "hello"
    msgstr "Привіт"
    """
    po_path = tmp_path / "uk.po"
    po_path.write_text(po_content, encoding="utf-8")

    provider = GettextProvider()
    await provider.load_file(po_path, "uk")

    assert provider.get("hello") == "Привіт"
    assert provider.get("missing", default="Fallback") == "Fallback"

    provider.set("hello", "Здоров")
    await provider.save_file()

    provider_new = GettextProvider()
    await provider_new.load_file(po_path, "uk")
    assert provider_new.get("hello") == "Здоров"



import pathlib

from vesna.meta.meta_default_object import MetaDefaultObject
from vesna.providers import ProviderProtocol


class Vesna(metaclass=MetaDefaultObject):
    """
    Provides a basic API for working with localisation, and also acts as a provider manager.
    """

    @MetaDefaultObject.meta_inject(auto_creation=True)
    def __init__(
        self, default_locale: str | None = None, default_path: pathlib.Path | str | None = None
    ) -> None:

        self.providers: dict[str, ProviderProtocol] = {}
        self.default_locale: str | None = default_locale
        self.default_path: pathlib.Path | str | None = default_path

    def _path_handler(self, path: pathlib.Path | str | None, locale_code: str) -> pathlib.Path:
        if isinstance(path, str):
            if "{locale_code}" in path:
                path: str = path.format(locale_code=locale_code)

            path: pathlib.Path = pathlib.Path(path)

        if not path or not path.is_file():
            raise FileNotFoundError(path)

        return path

    async def load_file(
        self,
        provider: ProviderProtocol | None = None,
        locale_code: str | None = None,
        path: pathlib.Path | str | None = None,
    ) -> None:
        locale_code = locale_code or self.default_locale
        if not locale_code:
            raise RuntimeError("Locale code is None")

        if provider:
            self.providers[locale_code] = provider

        provider = provider or self.providers.get(locale_code, None)
        if not provider:
            raise RuntimeError("Provider is None")

        if not provider.is_empty():
            raise RuntimeError("Provider is not empty")

        path = path or self.default_path
        if not path:
            raise RuntimeError("Path is None")
        path: pathlib.Path = self._path_handler(path, locale_code)

        await provider.load_file(path, locale_code)

    async def save_file(
        self, locale_code: str | None = None, path: pathlib.Path | str | None = None
    ) -> None:
        locale_code = locale_code or self.default_locale
        if not locale_code:
            raise RuntimeError("Locale code is None")

        path = path or self.default_path
        path: pathlib.Path = self._path_handler(path, locale_code)

        await self.providers[locale_code].save_file(path)

    def get_text(self, key: str, locale_code: str, is_exception: bool = False, **kwargs) -> str:
        """
        Searches for the value by key and language code.
        If there is no localisation code, it will try to search in default_locale.
        If there is no default_locale either, it will return an exception if is_exception is set,
        otherwise it will return the key.
        :param key: Key
        :param locale_code: Localisation code
        :param is_exception: If true, throws an exception; otherwise, returns the key.
        :param kwargs: Formatting arguments for the provider may not be supported by all providers.
        However, some providers (e.g. fluent) only support this type of formatting.
        :return: String
        """

        try:
            provider = self.providers.get(locale_code)
            if not provider:
                raise KeyError(f"Locale '{locale_code}' not loaded")
            value = provider.get(key, default=None, **kwargs)

            if value is None:
                raise KeyError(f"Key '{key}' missing")

            return value
        except KeyError as e:
            if self.default_locale and locale_code != self.default_locale:
                return self.get_text(key, self.default_locale, is_exception, **kwargs)

            if is_exception:
                raise KeyError(f"Key '{key}' not found in '{locale_code}'") from e
            return key

    def set_text(self, key: str, value: str, locale_code: str, is_exception: bool = False) -> str:
        """
        Sets the value by key. If there is no provider for the required localisation,
        it throws an exception if is_exception is true.
        :param key: Key
        :param value: Value
        :param locale_code: Localisation code
        :param is_exception: If true, throws an exception; otherwise, returns the key.
        :return: Value
        """

        provider = self.providers.get(locale_code)
        if not provider:
            if is_exception:
                raise KeyError(f"Locale '{locale_code}' not loaded")
            return key
        return provider.set(key, value)


vesna: Vesna = Vesna()

I18n = Vesna
i18n: I18n = vesna
Internationalization = Vesna
internationalization: Internationalization = vesna



import pathlib
import typing
from abc import ABC, abstractmethod

import pydantic


class DictProvider(ABC):
    """
    Almost complete abstract implementation of the provider protocol for formats that can be easily
    converted into a dictionary. Requires the implementation of IO methods.
    """

    @classmethod
    async def from_file(cls, path: pathlib.Path, locale_code: str) -> DictProvider:
        """
        Allows you to quickly create a new class instance and immediately load the localisation.
        :param path: Path
        :param locale_code: Localisation code
        :return: New instance
        """

        instance = cls()
        await instance.load_file(path, locale_code)
        return instance

    def __init__(self) -> None:
        self._storage: DataDictModel | None = None

    def __getitem__(self, key: str) -> str | None:
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> str:
        return self.set(key, value)

    def get(self, key: str, default: str | None = None, **kwargs) -> str | None:
        if self._storage is None:
            raise ValueError("The storage is None")

        parts = key.split(".")
        data = self._storage.data

        for part in parts:
            if isinstance(data, dict) and part in data:
                data = data.get(part)
            else:
                return default

        return str(data) if data is not None else default

    def set(self, key: str, value: str) -> str:
        if self._storage is None:
            raise ValueError("The storage is None")

        self._storage.data[key] = value
        return value

    def get_storage(self) -> DataDictModel:
        if self._storage is None:
            raise ValueError("The storage is None")

        return self._storage

    def get_locale_code(self) -> str | None:
        if self._storage is None:
            raise ValueError("The storage is None")

        return self._storage.locale_code

    def get_file_path(self) -> pathlib.Path | None:
        if self._storage is None:
            raise ValueError("The storage is None")

        return self._storage.path

    @abstractmethod
    async def load_file(self, path: pathlib.Path, locale_code: str) -> None:
        pass

    @abstractmethod
    async def save_file(self, path: pathlib.Path | None = None) -> None:
        pass

    async def clean(self) -> None:
        self._storage = None

    def is_empty(self) -> bool:
        return self._storage is None


class DataDictModel(pydantic.BaseModel):
    data: dict[str, typing.Any]
    locale_code: str | None  # ISO 15897 (ISO 639-1 + ISO 3166-1 alpha-2)
    path: pathlib.Path | None = None
    version: str = "1.0.0"



import pathlib

import aiofiles
import pydantic

try:
    import json5  # ty:ignore[unresolved-import, unused-ignore-comment, unused-ignore-comment]

    pyjson5 = None  # ty:ignore[unresolved-import, unused-ignore-comment, invalid-assignment]

    json5lib = json5
except ImportError:
    import pyjson5  # ty:ignore[unresolved-import, unused-ignore-comment, unused-ignore-comment]

    json5 = None  # ty:ignore[unresolved-import, unused-ignore-comment, invalid-assignment]

    json5lib = pyjson5

from vesna.providers.dict_provider import DataDictModel, DictProvider


class ProviderJSON5(DictProvider):
    """
    Implementation of IO methods for DictProvider to support the JSON5 format.
    Supports json5 or pyjson5 for the backend.
    """

    async def load_file(self, path: pathlib.Path, locale_code: str) -> None:
        await self.clean()

        async with aiofiles.open(path, "r", encoding="utf-8") as f:
            content = await f.read()
            raw_dict = json5lib.loads(content)
            try:
                self._storage = JSON5DataModel.model_validate(raw_dict)
            except pydantic.ValidationError:
                self._storage = JSON5DataModel(
                    data=raw_dict,
                    locale_code=locale_code,
                )

            self._storage.path = path

    async def save_file(self, path: pathlib.Path | None = None) -> None:
        if self._storage is None:
            raise ValueError("The storage is None")

        path = pathlib.Path(path) if path else self._storage.path
        if not path:
            raise ValueError(f"Path cannot be None | Path: {path}")

        data_to_save = self._storage.model_dump(exclude={"path"})

        async with aiofiles.open(path, "wb") as f:
            json5_string = json5lib.dumps(data_to_save)
            await f.write(json5_string.encode("utf-8"))


class JSON5DataModel(DataDictModel):
    pass



import pathlib

import aiofiles
import pydantic
from fluent.runtime import FluentBundle, FluentResource
from fluent.runtime.types import FluentNone


class FluentProvider:
    """
    Implementation of the protocol for the fluent format,
    does not support data recording and modification!
    """

    @classmethod
    async def from_file(cls, path: pathlib.Path, locale_code: str) -> FluentProvider:
        """
        Allows you to quickly create a new class instance and immediately load the localisation.
        :param path: Path
        :param locale_code: Localisation code
        :return: New instance
        """

        instance = cls()
        await instance.load_file(path, locale_code)
        return instance

    def __init__(self, use_isolating: bool = False) -> None:
        self._storage: FluentModel | None = None
        self.use_isolating = use_isolating

    def __getitem__(self, key: str) -> str | None | FluentNone:
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> str:
        return self.set(key, value)

    def get(self, key: str, default: str | None = None, **kwargs) -> str | None | FluentNone:
        """
        Returns the value by key, formats if kwargs is passed.
        :param key: Key
        :param default: If there is no key, the default will be returned.
        :param kwargs: Formatting arguments (if any).
        :return: Formatted string
        """
        if not self._storage:
            return default

        bundle = self._storage.data

        if not bundle.has_message(key):
            return default

        message = bundle.get_message(key)
        if message.value is None:
            return default

        formatted_string, errors = bundle.format_pattern(message.value, kwargs)
        if errors:
            raise RuntimeError(f"An error occurred while executing format_pattern: {errors}")

        return formatted_string

    def set(self, key: str, value: str) -> str:
        raise NotImplementedError("Fluent provider currently does not support runtime editing.")

    def get_storage(self) -> FluentModel:
        if self._storage is None:
            raise ValueError("The storage is None")
        return self._storage

    def get_locale_code(self) -> str | None:
        return self._storage.locale_code if self._storage else None

    def get_file_path(self) -> pathlib.Path | None:
        return self._storage.path if self._storage else None

    async def load_file(self, path: pathlib.Path, locale_code: str) -> None:
        await self.clean()

        async with aiofiles.open(path, "r", encoding="utf-8") as f:
            content = await f.read()

            resource = FluentResource(content)

            bundle = FluentBundle([locale_code], use_isolating=self.use_isolating)
            bundle.add_resource(resource)

            self._storage = FluentModel(data=bundle, path=path)

    async def save_file(self, path: pathlib.Path | None = None) -> None:
        raise NotImplementedError("Fluent provider currently does not support runtime saving.")

    async def clean(self) -> None:
        self._storage = None

    def is_empty(self) -> bool:
        return self._storage is None


class FluentModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    data: FluentBundle
    path: pathlib.Path | None = None

    version: str = "1.0.0"

    @property
    def locale_code(self) -> str | None:
        if self.data and hasattr(self.data, "locales") and self.data.locales:
            return self.data.locales[0]
        return None



import inspect
import typing
import weakref
from functools import wraps


class MetaDefaultObject(type):
    """
    A metaclass that provides basic DI tools can simulate global objects.
    """

    _meta_objects: weakref.WeakValueDictionary[type, object] = weakref.WeakValueDictionary()

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        cls.default_object = None

    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)

        if cls.default_object is None:
            cls.default_object = instance
            MetaDefaultObject._meta_objects[cls] = instance

        return instance

    def __setattr__(cls, name, value):
        super().__setattr__(name, value)

        if name == "default_object":
            if value is not None:
                MetaDefaultObject._meta_objects[cls] = value
            else:
                MetaDefaultObject._meta_objects.pop(cls, None)

    @classmethod
    def meta_get_instance(mcs, cls_type):
        return mcs._meta_objects.get(cls_type)

    @staticmethod
    def meta_inject(auto_creation: bool = False):
        def decorator(func: typing.Callable) -> typing.Any:
            sig = inspect.signature(func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                bound_args = sig.bind_partial(*args, **kwargs)

                for name, param in sig.parameters.items():
                    if name not in bound_args.arguments:
                        instance = MetaDefaultObject.meta_get_instance(param.annotation)

                        if instance:
                            bound_args.arguments[name] = instance
                        elif param.default is inspect.Parameter.empty:
                            if auto_creation and inspect.isclass(param.annotation):
                                instance = param.annotation()
                                bound_args.arguments[name] = instance
                            else:
                                raise TypeError(
                                    f"Missing required argument: '{name}' ({param.annotation}). \
                                No instance found in registry."
                                )

                bound_args.apply_defaults()
                return func(*bound_args.args, **bound_args.kwargs)

            return wrapper

        return decorator



from .protocol import ModelDataProtocol, ProviderProtocol

__all__ = ["ProviderProtocol", "ModelDataProtocol"]

try:
    from .yaml import ProviderYAML  # noqa: F401

    __all__.append("ProviderYAML")
except ImportError:
    pass

try:
    from .json import ProviderJSON  # noqa: F401

    __all__.append("ProviderJSON")
except ImportError:
    pass

try:
    from .hjson import ProviderHJSON  # noqa: F401

    __all__.append("ProviderHJSON")
except ImportError:
    pass

try:
    from .json5 import ProviderJSON5  # noqa: F401

    __all__.append("ProviderJSON5")
except ImportError:
    pass

try:
    from .toml import ProviderTOML  # noqa: F401

    __all__.append("ProviderTOML")
except ImportError:
    pass

try:
    from .gettext_provider import GettextProvider  # noqa: F401

    __all__.append("GettextProvider")
except ImportError:
    pass

try:
    from .fluent_provider import FluentProvider  # noqa: F401

    __all__.append("FluentProvider")
except ImportError:
    pass



import pytest

import vesna


@pytest.fixture(autouse=True)
async def setup_i18n():
    try:
        await vesna.i18n.load_file(
            vesna.providers.ProviderJSON(), "uk", "tests/localisation/{locale_code}.json"
        )
    except RuntimeError, KeyError, FileNotFoundError:
        pytest.fail("Failed to load translation file")


@pytest.fixture
def locale():
    return vesna.Locale("uk")


async def test_load():
    provider = vesna.providers.ProviderJSON()
    locale_code = "en"
    path = "tests/localisation/{locale_code}.json"
    i18n = vesna.Vesna()

    with pytest.raises(RuntimeError):
        await i18n.load_file(None, locale_code, path)

    with pytest.raises(RuntimeError):
        await i18n.load_file(provider, None, path)

    with pytest.raises(RuntimeError):
        await i18n.load_file(provider, locale_code, None)

    await i18n.load_file(provider, locale_code, path)


async def test_save():
    with pytest.raises(FileNotFoundError):
        await vesna.i18n.save_file("en")

    with pytest.raises(RuntimeError):
        await vesna.i18n.save_file(None)


def test_get(locale):
    assert locale.get("hello") == "Привіт {}"


def test_key(locale):
    assert locale["hello"] == "Привіт {}"



import datetime
import decimal

import babel.dates
import babel.numbers
from babel import Locale

from vesna.meta.meta_default_object import MetaDefaultObject


class BabelHandler(metaclass=MetaDefaultObject):
    """
    Provides basic tools for working with Babel.
    """

    def __init__(self):
        self._locale_cache: dict[str, Locale] = {}

    def get_locale(self, locale_code: str) -> Locale:
        if locale_code not in self._locale_cache:
            self._locale_cache[locale_code] = Locale.parse(locale_code)
        return self._locale_cache[locale_code]

    def format_date(
        self,
        date: datetime.date | None,
        locale_code: str,
        format: str = "medium",
    ):
        return babel.dates.format_date(date, format=format, locale=locale_code)

    def format_currency(
        self, amount: float | decimal.Decimal | str, currency: str, locale_code: str
    ) -> str:
        return babel.numbers.format_currency(amount, currency, locale=locale_code)

    def get_plural_form(self, count: int | float | decimal.Decimal, locale_code: str) -> str:
        locale = self.get_locale(locale_code)
        return locale.plural_form(count)

    def format_decimal(self, value: float | int, locale_code: str) -> str:
        return babel.numbers.format_decimal(value, locale=locale_code)



import datetime
import decimal

try:
    import babel

    from vesna.babel_handler import BabelHandler

    SUPPORT_BABEL = True
except ImportError:
    babel = None  # ty:ignore[invalid-assignment]
    BabelHandler = None  # ty:ignore[invalid-assignment]

    SUPPORT_BABEL = False

from vesna.meta.meta_cache import MetaCache
from vesna.vesna import Vesna


class Locale(metaclass=MetaCache):
    """
    Provides a basic API for working with a single localisation.
    """

    @Vesna.meta_inject(auto_creation=True)
    def __init__(self, locale_code: str, vesna: Vesna, babel_handler: BabelHandler) -> None:
        self.locale_code: str = locale_code
        self.vesna: Vesna = vesna
        self.babel_handler: BabelHandler = babel_handler

    def __getitem__(self, key: str) -> str:
        return self.vesna.get_text(key, self.locale_code)

    def __call__(self, key: str, *args, **kwargs) -> str:
        return self.format(key, *args, **kwargs)

    def format(self, key: str, *args, **kwargs) -> str:
        text = self.vesna.get_text(key, self.locale_code)
        return text.format(*args, **kwargs) if (args or kwargs) else text

    def get(self, key: str, **kwargs) -> str:
        return self.vesna.get_text(key, self.locale_code, **kwargs)

    if SUPPORT_BABEL:

        @property
        def babel(self) -> babel.Locale:  # ty:ignore[unresolved-attribute]
            return self.babel_handler.get_locale(self.locale_code)

        def date(self, value: datetime.datetime | None, format: str = "medium") -> str:
            return self.babel_handler.format_date(value, self.locale_code, format)

        def plural(self, key: str, count: int, **kwargs) -> str:
            plural_form = self.babel_handler.get_plural_form(count, self.locale_code)

            full_key = f"{key}.{plural_form}"
            text = self.get(full_key)

            if text == full_key:
                text = self.get(key)

            return text.format(count=count, **kwargs)

        def currency(
            self, amount: float | decimal.Decimal | str, currency_code: str = "USD"
        ) -> str:
            return self.babel_handler.format_currency(amount, currency_code, self.locale_code)

        def number(self, value: float | int) -> str:
            return self.babel_handler.format_decimal(value, self.locale_code)



import typing
from weakref import WeakValueDictionary


class MetaCache(type):
    """
    A metaclass that provides object caching
    if identical arguments are passed when creating an object.
    """

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        cls._meta_cache: WeakValueDictionary[typing.Any, typing.Any] = WeakValueDictionary()

    def __call__(cls, *args, **kwargs):
        cache_key = (args, tuple(sorted(kwargs.items())))
        instance = cls._meta_cache.get(cache_key)
        if instance is None:
            instance = super().__call__(*args, **kwargs)
            cls._meta_cache[cache_key] = instance

        return instance



import pathlib
from io import BytesIO, StringIO

import aiofiles
import pydantic
from babel.messages.catalog import Catalog
from babel.messages.mofile import read_mo, write_mo
from babel.messages.pofile import read_po, write_po


class GettextProvider:
    """
    Implementation of the protocol for the GNU Gettext format.
    """

    @classmethod
    async def from_file(cls, path: pathlib.Path, locale_code: str) -> GettextProvider:
        """
        Allows you to quickly create a new class instance and immediately load the localisation.
        :param path: Path
        :param locale_code: Localisation code
        :return: New instance
        """

        instance = cls()
        await instance.load_file(path, locale_code)
        return instance

    def __init__(self) -> None:
        self._storage: GettextModel | None = None

    def __getitem__(self, key: str) -> str | None:
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> str:
        return self.set(key, value)

    def get(self, key: str, default: str | None = None, **kwargs) -> str | None:
        if not self._storage:
            return default

        message = self._storage.data.get(key)

        if message and message.string:
            if isinstance(message.string, str):
                return message.string

            elif isinstance(message.string, tuple):
                return message.string[0]

        return default

    def set(self, key: str, value: str) -> str:
        if not self._storage:
            raise RuntimeError("Storage is empty")

        if key in self._storage.data:
            self._storage.data[key].string = value
        else:
            self._storage.data.add(key, string=value)

        return value

    def get_storage(self) -> GettextModel:
        if self._storage is None:
            raise ValueError("The storage is None")
        return self._storage

    def get_locale_code(self) -> str | None:
        return self._storage.locale_code if self._storage else None

    def get_file_path(self) -> pathlib.Path | None:
        return self._storage.path if self._storage else None

    async def load_file(self, path: pathlib.Path, locale_code: str) -> None:
        """
        Supports loading po and mo files

        :param path: File path
        :param locale_code: Locale code
        :return: None
        """

        await self.clean()

        if path.suffix == ".mo":
            # Читаємо бінарний MO файл
            async with aiofiles.open(path, "rb") as f:
                content = await f.read()

                buf = BytesIO(content)
                catalog = read_mo(buf)
                catalog.locale = locale_code
        else:
            async with aiofiles.open(path, "r", encoding="utf-8") as f:
                content = await f.read()

                buf = StringIO(content)
                catalog = read_po(buf, locale=locale_code)

        self._storage = GettextModel(data=catalog, path=path)

    async def save_file(self, path: pathlib.Path | None = None) -> None:
        """
        Supports saving in po format

        :param path: The path to the file to be saved;
        if not specified, it is saved in the same folder from which it was loaded.
        :return: None
        """

        target_path = path or (self._storage.path if self._storage else None)
        if not target_path or not self._storage:
            raise ValueError("Path or storage is empty")

        buf = BytesIO()
        write_po(buf, self._storage.data)

        async with aiofiles.open(target_path, "wb") as f:
            await f.write(buf.getvalue())

    async def compile_mo(self, path: pathlib.Path | None = None) -> None:
        """
        Supports compile in mo format

        :param path: The path to the file to be saved;
        if not specified, it is saved in the same folder from which it was loaded.
        :return: None
        """

        if not self._storage:
            raise RuntimeError("Storage is empty, nothing to compile")

        target_path = path
        if not target_path:
            if self._storage.path:
                target_path = self._storage.path.with_suffix(".mo")
            else:
                raise ValueError("Path is empty")

        buf = BytesIO()
        write_mo(buf, self._storage.data)

        async with aiofiles.open(target_path, "wb") as f:
            await f.write(buf.getvalue())

    async def clean(self) -> None:
        self._storage = None

    def is_empty(self) -> bool:
        return self._storage is None


class GettextModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    data: Catalog
    path: pathlib.Path | None = None

    version: str = "1.0.0"

    @property
    def locale_code(self) -> str | None:
        return str(self.data.locale) if self.data.locale else None
